@@ -794,19 +794,15 @@ remote_wait_for_cancel(ProxyFunction *func)
 	ProxyConnection *conn;
 	ProxyCluster *cluster = func->cur_cluster;
 	int			i,
-				pending = 1;
+				pending;
 	struct timeval now;
 
 	/* now loop until all results are arrived */
-	while (pending)
+	while (1)
 	{
 		/* allow postgres to cancel processing */
 		CHECK_FOR_INTERRUPTS();
 
-		/* wait for events */
-		if (poll_conns(func, cluster) == 0)
-			continue;
-
 		/* recheck */
 		pending = 0;
 		gettimeofday(&now, NULL);
@@ -816,10 +812,15 @@ remote_wait_for_cancel(ProxyFunction *func)
 			if (!conn->run_tag)
 				continue;
 
-			if (conn->cur->state != C_DONE)
+			if (conn->cur->state == C_QUERY_READ)
 				pending++;
 			check_timeouts(func, cluster, conn, now.tv_sec);
 		}
+		if (!pending)
+			break;
+
+		/* wait for events */
+		poll_conns(func, cluster);
 	}
 
 	/* review results, calculate total */
@@ -830,7 +831,7 @@ remote_wait_for_cancel(ProxyFunction *func)
 		if (!conn->run_tag)
 			continue;
 
-		if (conn->cur->state != C_DONE)
+		if (conn->cur->state != C_DONE && conn->cur->state != C_NONE)
 			plproxy_error(func, "Unfinished connection: %d", conn->cur->state);
 		if (conn->res != NULL)
 		{
