@@ -71,6 +71,33 @@ static bool usable_binary(Oid oid)
 	}
 }
 
+bool
+plproxy_composite_valid(ProxyComposite *type)
+{
+	HeapTuple type_tuple;
+	HeapTuple rel_tuple;
+	Form_pg_type pg_type;
+	Oid oid = type->tupdesc->tdtypeid;
+	bool res;
+
+	if (!type->alterable)
+		return true;
+	type_tuple = SearchSysCache(TYPEOID, ObjectIdGetDatum(oid), 0, 0, 0);
+	if (!HeapTupleIsValid(type_tuple))
+		elog(ERROR, "cache lookup failed for type %u", oid);
+
+	pg_type = (Form_pg_type) GETSTRUCT(type_tuple);
+	rel_tuple = SearchSysCache(RELOID, ObjectIdGetDatum(pg_type->typrelid), 0, 0, 0);
+	if (!HeapTupleIsValid(rel_tuple))
+		elog(ERROR, "cache lookup failed for type relation %u", pg_type->typrelid);
+
+	res = plproxy_check_stamp(&type->stamp, rel_tuple);
+
+	ReleaseSysCache(rel_tuple);
+	ReleaseSysCache(type_tuple);
+
+	return res;
+}
 /*
  * Collects info about fields of a composite type.
  *
@@ -86,6 +113,7 @@ plproxy_composite_info(ProxyFunction *func, TupleDesc tupdesc)
 	Form_pg_attribute a;
 	ProxyType  *type;
 	const char *name;
+	Oid oid = tupdesc->tdtypeid;
 
 	old_ctx = MemoryContextSwitchTo(func->ctx);
 
@@ -95,13 +123,42 @@ plproxy_composite_info(ProxyFunction *func, TupleDesc tupdesc)
 	ret->tupdesc = BlessTupleDesc(tupdesc);
 	ret->use_binary = 1;
 
+	ret->alterable = 0;
+	if (oid != RECORDOID)
+	{
+		HeapTuple type_tuple;
+		HeapTuple rel_tuple;
+		Form_pg_type pg_type;
+
+		type_tuple = SearchSysCache(TYPEOID, ObjectIdGetDatum(oid), 0, 0, 0);
+		if (!HeapTupleIsValid(type_tuple))
+			elog(ERROR, "cache lookup failed for type %u", oid);
+		pg_type = (Form_pg_type) GETSTRUCT(type_tuple);
+		rel_tuple = SearchSysCache(RELOID, ObjectIdGetDatum(pg_type->typrelid), 0, 0, 0);
+		if (!HeapTupleIsValid(rel_tuple))
+			elog(ERROR, "cache lookup failed for type relation %u", pg_type->typrelid);
+		plproxy_set_stamp(&ret->stamp, rel_tuple);
+		ReleaseSysCache(rel_tuple);
+		ReleaseSysCache(type_tuple);
+		ret->alterable = 1;
+
+		if (ret->tupdesc->tdtypeid != oid)
+			elog(ERROR, "lost oid");
+	}
+
 	MemoryContextSwitchTo(old_ctx);
 
+	ret->nfields = 0;
 	for (i = 0; i < natts; i++)
 	{
 		a = tupdesc->attrs[i];
 		if (a->attisdropped)
-			plproxy_error(func, "dropped attrs not supported");
+		{
+			ret->name_list[i] = NULL;
+			ret->type_list[i] = NULL;
+			continue;
+		}
+		ret->nfields++;
 
 		name = quote_identifier(NameStr(a->attname));
 		ret->name_list[i] = plproxy_func_strdup(func, name);
@@ -125,7 +182,8 @@ plproxy_free_composite(ProxyComposite *rec)
 	for (i = 0; i < natts; i++)
 	{
 		plproxy_free_type(rec->type_list[i]);
-		pfree(rec->name_list[i]);
+		if (rec->name_list[i])
+			pfree(rec->name_list[i]);
 	}
 	pfree(rec->type_list);
 	pfree(rec->name_list);
@@ -136,6 +194,9 @@ plproxy_free_composite(ProxyComposite *rec)
 void
 plproxy_free_type(ProxyType *type)
 {
+	if (type == NULL)
+		return;
+
 	if (type->name)
 		pfree(type->name);
 
@@ -169,7 +230,11 @@ plproxy_recv_composite(ProxyComposite *meta, char **values, int *lengths, int *f
 	for (i = 0; i < natts; i++)
 	{
 		if (tupdesc->attrs[i]->attisdropped)
-			elog(ERROR, "dropped attrs not supported");
+		{
+			dvalues[i] = (Datum)NULL;
+			nulls[i] = 'n';
+			continue;
+		}
 
 		dvalues[i] = plproxy_recv_type(meta->type_list[i],
 									   values[i], lengths[i], fmts[i]);
@@ -237,14 +302,18 @@ plproxy_find_type_info(ProxyFunction *func, Oid oid, bool for_send)
 	switch (s_type->typtype)
 	{
 		default:
-		case 'p':
+			plproxy_error(func, "unsupported type code: %s (%u)", namebuf, oid);
+			break;
+		case TYPTYPE_PSEUDO:
 			if (oid != VOIDOID)
 				plproxy_error(func, "unsupported pseudo type: %s (%u)",
 							  namebuf, oid);
-		case 'b':
-		case 'c':
-		case 'd':
-		case 'e':
+			break;
+		case TYPTYPE_BASE:
+		case TYPTYPE_COMPOSITE:
+		case TYPTYPE_DOMAIN:
+		case TYPTYPE_ENUM:
+		case TYPTYPE_RANGE:
 			break;
 	}
 
